;;; sscell.el --- Attach notes to messages    -*- lexical-binding: t -*-

;; Copyright (C) 2017 Free Software Foundation, Inc

;; Author: Michael Heerdegen <michael_heerdegen@web.de>
;; Maintainer: Michael Heerdegen <michael_heerdegen@web.de>
;; Created: 2017_12_11
;; Keywords: lisp
;; Version: 0.1
;; Package-Requires: ()

(defvar sscell--tag (make-symbol "sscell"))

(defun sscell-p (object)
  (eq (car-safe object)  sscell--tag))

(defmacro sscell-make (deps &rest body)
  "Make an sscell.
DEPS is a list of the dependencies of the sscell.
A dependency is either:
  EXPR
or
  (:test TESTFUN EXPR)

The return value is an sscell object..."
  (cl-callf or deps '(t))
  (let ((last-result      (make-symbol "last-result"))
        (last-dep-results (make-symbol "last-dep-results"))
        (new-dep-results  (make-symbol "new-dep-results"))
        (get-deps         (make-symbol "get-deps"))
        (instruction      (make-symbol "instruction"))
        (test-validity-p  (make-symbol "test-validity-p"))
        (old              (make-symbol "old"))
        (new              (make-symbol "new"))
        (tests (cl-maplist
                (lambda (more-deps) (let ((dep (car more-deps)))
                                 (if (not (eq (car-safe dep) :test))
                                     '#'eq
                                   (prog1 (nth 1 dep)
                                     (setcar more-deps (nth 2 dep))))))
                deps)))
    `(let* (,last-result
            (,last-dep-results nil)
            (,get-deps (lambda () (list ,@deps)))
            (,test-validity-p (lambda (,old ,new)
                                (and ,last-dep-results
                                     (cl-every #'identity
                                               (cl-mapcar #'funcall (list ,@tests) ,old ,new))))))
       (cons sscell--tag
             (lambda (,instruction)
               (pcase ,instruction
                 ('test (and ,last-dep-results
                             (funcall ,test-validity-p ,last-dep-results (funcall ,get-deps))))
                 ('get (let ((,new-dep-results (funcall ,get-deps)))
                         (if (funcall ,test-validity-p ,last-dep-results ,new-dep-results)
                             ,last-result
                           (setq ,last-dep-results ,new-dep-results
                                 ,last-result      (progn ,@body)))))))))))

(defun sscell-get (sscell)
  (cl-assert (sscell-p sscell))
  (funcall (cdr sscell) 'get))

(defun sscell-valid-p (sscell)
  (cl-assert (sscell-p sscell))
  (funcall (cdr sscell) 'test))

(defmacro sscell-let (bindings &rest body)
  (declare (indent 1) (debug fixme))
  (cl-callf2 mapcar
      (pcase-lambda (`(,var ,deps ,binding))
        (list (make-symbol (concat (symbol-name var) "-sscell"))
              var deps binding))
      bindings)
  `(let ,(mapcar
          (pcase-lambda (`(,helper-var ,_var ,deps ,binding))
            `(,helper-var (sscell-make ,deps ,binding)))
          bindings)
     (cl-symbol-macrolet
         ,(mapcar (pcase-lambda (`(,helper-var ,var ,_deps ,_binding))
                    `(,var (sscell-get ,helper-var)))
                  bindings)
       ,@body)))

(defmacro sscell-let* (bindings &rest body)
  (declare (indent 1) (debug fixme))
  (cl-reduce
   (lambda (expr binding) `(sscell-let (,binding) ,expr))
   (nreverse bindings)
   :initial-value (macroexp-progn body)))
